#!/usr/bin/python3

import ctypes
import os
import time
import re

class FTDI_PatternGenerator:

   def __init__(self, FTDI_PatternGenerator_Folder_Path):
      FTDI_PatternGenerator_DLL_Path = os.path.join(FTDI_PatternGenerator_Folder_Path, 'FTDI_Pattern_Generator.dll')
      FTDI_PatternGenerator_ErrorCSV_Path = os.path.join(FTDI_PatternGenerator_Folder_Path, 'FTDI_Pattern_Errors.csv')
      
      self.FTDI_PatternGenerator_DLL = ctypes.cdll.LoadLibrary(FTDI_PatternGenerator_DLL_Path)
      
      self.FTDI_ErrorDict = {}
      with open(FTDI_PatternGenerator_ErrorCSV_Path) as ErrorFile:
         ErrorList = ErrorFile.read().splitlines()
      ErrorList.pop(0) #removing the header
      for item in ErrorList:
         ErrorItem = item.split(',', 1)
         self.FTDI_ErrorDict[int(ErrorItem[0])] = ErrorItem[1]
         
      self.FTDIDeviceLUT = [] #initializing with empty values
      
   def __GetTargetDict(self, InputDictList, TargetKey, TargetValue, DefaultValue):#Private Method
      TargetDict = DefaultValue
      for item in InputDictList:
          if item[TargetKey] == TargetValue:
              TargetDict = item
              break
         
      return TargetDict
   
   def __HandleFTDIError(self, Error_Status):#Private Method
      try:
         raise Exception(self.FTDI_ErrorDict[Error_Status])
      except KeyError: #Key does not occur
         raise Exception("UNKNOWN_ERROR, Unknown error occured")
      
   
   def InitializeDevice(self, DeviceName, PortName, BaudRate, PatterFilePath):
      EVM_Index = ctypes.c_int(-1)
      DeviceIndex = ctypes.c_int(-1)
      Error_Status = 0
      
      CurrentFTDIDevice = self.__GetTargetDict(self.FTDIDeviceLUT, 'PortName', PortName, False) #checking if the port already opened
      if(CurrentFTDIDevice == False):
         Error_Status = self.FTDI_PatternGenerator_DLL.openUSBPort(PortName.encode('ascii', 'ignore'),ctypes.c_ulong(BaudRate),ctypes.byref(EVM_Index))
         if (Error_Status != 0):
            self.__HandleFTDIError(Error_Status)            
      else:
         EVM_Index = CurrentFTDIDevice["EVM_Index"]
         
      CurrentFTDIDevice = self.__GetTargetDict(self.FTDIDeviceLUT, 'DeviceName', DeviceName, False) #checking if the DeviceName already used
      if(CurrentFTDIDevice == False):
         Error_Status = self.FTDI_PatternGenerator_DLL.initializeDevice(PatterFilePath.encode('ascii', 'ignore'),EVM_Index,ctypes.byref(DeviceIndex))
         if (Error_Status != 0):
            self.__HandleFTDIError(Error_Status)
      else:
         raise Exception("DEVICE_ALREADY_INITIALIZED, With the current DeviceName {0}, a Device is already Initialized".format(DeviceName))

      self.FTDIDeviceLUT.append({"DeviceName":DeviceName, "PortName":PortName, "BaudRate":BaudRate, "PatterFilePath":PatterFilePath, "EVM_Index":EVM_Index,"DeviceIndex":DeviceIndex})
     

   def WriteRegister(self, DeviceName, WriteAddr, WriteData):
      Error_Status = 0
      
      CurrentFTDIDevice = self.__GetTargetDict(self.FTDIDeviceLUT, 'DeviceName', DeviceName, False) #checking if the DeviceName already exists

      if(CurrentFTDIDevice == False):
         raise Exception("DEVICE_NOT_INITIALIZED, Target Device {0} is Not Initialized".format(DeviceName))
      
      Error_Status = self.FTDI_PatternGenerator_DLL.writeRegister(CurrentFTDIDevice["EVM_Index"],CurrentFTDIDevice["DeviceIndex"],ctypes.c_ulong(WriteAddr), ctypes.c_ulong(WriteData))
      
      if (Error_Status != 0):
         self.__HandleFTDIError(Error_Status)
      
   def ReadRegister(self, DeviceName, ReadAddr):
      Error_Status = 0
      ReadData = ctypes.c_ulong(0)
      
      CurrentFTDIDevice = self.__GetTargetDict(self.FTDIDeviceLUT, 'DeviceName', DeviceName, False) #checking if the DeviceName already used
      
      if(CurrentFTDIDevice == False):
         raise Exception("DEVICE_NOT_INITIALIZED, Target Device {0} is Not Initialized".format(DeviceName))
      
      Error_Status = self.FTDI_PatternGenerator_DLL.readRegister(CurrentFTDIDevice["EVM_Index"],CurrentFTDIDevice["DeviceIndex"],ctypes.c_ulong(ReadAddr), ctypes.byref(ReadData))
      
      if (Error_Status != 0):
         self.__HandleFTDIError(Error_Status)
      
      return ReadData.value
   
   def CloseAllDev(self):
      Error_Status = 0
      
      Error_Status = self.FTDI_PatternGenerator_DLL.closeAllPorts()
      
      if (Error_Status != 0):
         self.__HandleFTDIError(Error_Status)
      
      self.FTDIDeviceLUT = [] #re-initializing with empty values
      
      
   def WriteMultipleRegisters(self, DeviceName, WriteAddrList, WriteDataList):
      Error_Status = 0
      
      CurrentFTDIDevice = self.__GetTargetDict(self.FTDIDeviceLUT, 'DeviceName', DeviceName, False) #checking if the DeviceName already used

      if(CurrentFTDIDevice == False):
         raise Exception("DEVICE_NOT_INITIALIZED, Target Device {0} is Not Initialized".format(DeviceName))
      
      WriteAddrArray = (ctypes.c_ulong * len(WriteAddrList))(*WriteAddrList)
      WriteDataArray = (ctypes.c_ulong * len(WriteDataList))(*WriteDataList)
      
      Error_Status = self.FTDI_PatternGenerator_DLL.writeRegister_Array(CurrentFTDIDevice["EVM_Index"],CurrentFTDIDevice["DeviceIndex"], WriteAddrArray, WriteDataArray, ctypes.c_int(len(WriteAddrList)))
      
      if (Error_Status != 0):
         self.__HandleFTDIError(Error_Status)
      
   def ReadMultipleRegisters(self, DeviceName, ReadAddrList):
      Error_Status = 0
      
      CurrentFTDIDevice = self.__GetTargetDict(self.FTDIDeviceLUT, 'DeviceName', DeviceName, False) #checking if the DeviceName already used

      if(CurrentFTDIDevice == False):
         raise Exception("DEVICE_NOT_INITIALIZED, Target Device {0} is Not Initialized".format(DeviceName))
      
      ReadAddrArray = (ctypes.c_ulong * len(ReadAddrList))(*ReadAddrList)
      ReadDataArray = (ctypes.c_ulong * len(ReadAddrList))()
      
      Error_Status = self.FTDI_PatternGenerator_DLL.readRegister_Array(CurrentFTDIDevice["EVM_Index"],CurrentFTDIDevice["DeviceIndex"], ReadAddrArray, ReadDataArray, ctypes.c_int(len(ReadAddrList)))
      if (Error_Status != 0):
         self.__HandleFTDIError(Error_Status)
      
      return list(ReadDataArray)
   
   def LoadConfigFile(self, ConfigFilePath):
      ConfigOperationDetail = []#initializing with empty values
      
      try:
         CurrentDeviceName = self.FTDIDeviceLUT[0]['DeviceName']
      except:
         raise Exception("DEVICE_NOT_INITIALIZED, Target Device is Not Initialized")
      
      #Parsing the configuration
      with open(ConfigFilePath) as ConfigFile:
          ConfigList = ConfigFile.read().splitlines()

      Configline = 0    
      for ConfigItem in ConfigList:
         Configline+= 1#incrementing
         
         if (ConfigItem[:2] == r"//") or (ConfigItem[:2] == r"\\") or (re.match( r'^comment', ConfigItem, re.I) != None):
             ConfigOperation = "Comment"
             continue
            
         elif (re.match( r'^delay', ConfigItem, re.I) != None):
             ErrorFlag = 0
             ConfigOperation = "Delay"
             matchObj = re.match( r'^delay\((.*?)\)', ConfigItem, re.I)
             try:
                 DelayInSec = float(matchObj.group(1))
             except:
                 DelayInSec = 0
                 raise Exception("Delay command mentioned at line {0} of {1} is not in expected format, expected delay command format is \ndelay(5) \\\\ to introduce a delay of 5 second".format(Configline, ConfigFilePath))
             ConfigOperationDetail.append({'ConfigOperation' : ConfigOperation, 'DeviceName' : CurrentDeviceName, 'RegisterAddr' : 0, 'RegisterValue' : 0, 'DelayInSec' : DelayInSec})
                
         elif (re.match( r'^0?x', ConfigItem, re.I) != None):
            ErrorFlag = 0
            ConfigOperation = "Write"
            matchObj = re.split( r'[\s\t]+', ConfigItem)
            try:
               RegisterAddr = int(matchObj[0], 16) #16 -> base of hex to int conversion
            except ValueError:
               RegisterAddr = 0
               ErrorFlag = 1
            try:
               RegisterValue = int(matchObj[1], 16) #16 -> base of hex to int conversion
            except ValueError:
               RegisterValue = 0
               ErrorFlag = 1
                
            if (ErrorFlag == 0):
               ConfigOperationDetail.append({'ConfigOperation' : ConfigOperation, 'DeviceName' : CurrentDeviceName, 'RegisterAddr' : RegisterAddr, 'RegisterValue' : RegisterValue, 'DelayInSec' : 0})
            else:
               raise Exception("Register Write command mentioned at line {0} of {1} is not in expected format, expected register write command format is \n0x05 0x25 \\\\ to write 0x25 value to 0x05 register".format(Configline, ConfigFilePath))
         else:
             #New DeviceName?
             if (len(ConfigItem)>0):
                TempDeviceName = ConfigItem.split()[0]
                if(self.__GetTargetDict(self.FTDIDeviceLUT, 'DeviceName', TempDeviceName, False) != False): #checking if the DeviceName already used
                   CurrentDeviceName = TempDeviceName
                else:
                   raise Exception("{0} is an invalid DeviceName mentioned at line {1} of {2}. Initialize the {0} as a new device, if required".format(TempDeviceName, Configline, ConfigFilePath))
                     
      #Executing the configuration
                     
      for EachConfigDetail in ConfigOperationDetail:
         if (EachConfigDetail['ConfigOperation'] == "Write"):
            self.WriteRegister(EachConfigDetail['DeviceName'], EachConfigDetail['RegisterAddr'], EachConfigDetail['RegisterValue'])
         elif (EachConfigDetail['ConfigOperation'] == "Delay"):
            time.sleep(EachConfigDetail['DelayInSec'])# Delay in Seconds
            
